/* ----------------------------------------------------------------------------
 * Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
 * ON Semiconductor), All Rights Reserved
 *
 * This code is the property of ON Semiconductor and may not be redistributed
 * in any form without prior written permission from ON Semiconductor.
 * The terms of use and warranty for this code are covered by contractual
 * agreements between ON Semiconductor and the licensee.
 *
 * This is Reusable Code.
 *
 * ------------------------------------------------------------------------- */

#include <string.h>

#include "codecs/base/baseCodec.h"
#include "codecs/codecInternal.h"

/* Define function prototypes for the features we will override from the base */
static void echoEncode(CODEC codec);

static void echoDecode(CODEC codec);

/* ----------------------------------------------------------------------------
 * Function     : initialiseStructure
 * ----------------------------------------------------------------------------
 * Description  : Initialises an echo codec control block. This sets the
 *                function pointers in the jump table for any methods we
 *                wish to override.
 * Inputs       : structure : A pointer to a codec control structure which
 *                is to be initialised
 * Outputs      : The structure object is initialised with default data.
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void initialiseStructure(pCodec structure)
{
    if (codecIsCodec(structure))
    {
        structure->type = ECHO;
        structure->encode = &echoEncode;
        structure->decode = &echoDecode;
    }
}

/* ----------------------------------------------------------------------------
 * Function     : makeEchoCodec
 * ----------------------------------------------------------------------------
 * Description  : General constructor for an echo codec, this uses malloc to
 *                allocate a block of memory which it then initialises as
 *                above.
 * Inputs       : None
 * Outputs      : An opaque CODEC object, effectively a handle that van be
 *                used in later operations on this codec
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
CODEC makeEchoCodec(void)
{
    /* "derive" from the base codec */
    pCodec echo = getCodec(makeBaseCodec());
    initialiseStructure(echo);
    return echo;
}

/* ----------------------------------------------------------------------------
 * Function     : populateEchoCodec
 * ----------------------------------------------------------------------------
 * Description  : As an alternative to allocating memory for a codec, this
 *                method is provided to allow a pre-allocated memory
 *                location to be set up
 * Inputs       : buffer : a pointer to a buffer in which to create and
 *                initialise the codec control block
 *                size : the size in bytes of the buffer
 * Outputs      : NULL if the size of the buffer is not large enough or a
 *                CODEC handle
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
CODEC populateEchoCodec(unsigned char *buffer, uint32_t size)
{
    /* "derive" from the base codec */
    pCodec echo = getCodec(populateBaseCodec(buffer, size));
    initialiseStructure(echo);
    return buffer;
}

/* ----------------------------------------------------------------------------
 * Function     : echoEncode
 * ----------------------------------------------------------------------------
 * Description  : The echo codec simply copies the input buffer to the
 *                output buffer and is provided as a minimum codec
 *                application
 * Inputs       : codec : An opaque handle to a codec control block
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void echoEncode(CODEC codec)
{
    pCodec echo = getCodec(codec);
    uint32_t sizeToEcho =
        (echo->inputBufferSize > echo->outputBufferMaxSize) ?
        echo->outputBufferMaxSize : echo->inputBufferSize;
    memcpy((void *)echo->outputBufferAddress,
           (void *)echo->inputBufferAddress, sizeToEcho);
    echo->outputBufferUsedSize = sizeToEcho;
    echo->notifyComplete();
}

/* ----------------------------------------------------------------------------
 * Function     : echoDecode
 * ----------------------------------------------------------------------------
 * Description  : The echo codec simply copies the input buffer to the
 *                output buffer and is provided as a minimum codec
 *                application. Specifically the decode operation is the same
 *                as the encode, so just call it
 * Inputs       : codec : An opaque handle to a codec control block
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void echoDecode(CODEC codec)
{
    /* encode and decode will do the same thing when echoing */
    echoEncode(codec);
}
