/* ----------------------------------------------------------------------------
 * Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
 * ON Semiconductor), All Rights Reserved
 *
 * This code is the property of ON Semiconductor and may not be redistributed
 * in any form without prior written permission from ON Semiconductor.
 * The terms of use and warranty for this code are covered by contractual
 * agreements between ON Semiconductor and the licensee.
 *
 * This is Reusable Code.
 *
 * ------------------------------------------------------------------------- */

#include <malloc.h>

#include "logger.h"

#include "io/base/ioBase.h"
#include "ioEcho.h"

static uint32_t readCounter = 0;
static uint32_t writeCounter = 0;

/* Define function prototypes for the features we will override from the base */
static void echoRead(pSourceSink source, unsigned char *buffer,
                     uint32_t maxsize);

static void echoWrite(pSourceSink sink, unsigned char *buffer, uint32_t size);

/* ----------------------------------------------------------------------------
 * Function     : makeEchoSourceSink
 * ----------------------------------------------------------------------------
 * Description  : This IO provider allows test data to be input and output
 *                to the codec and is provided for debug purposes. This
 *                method allows the construction of a suitable control block
 *                using malloc.
 * Inputs       : None
 * Outputs      : Pointer to an IO Source/Sink control block which has been
 *                configured to match the test data indicated above.
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
pSourceSink makeEchoSourceSink(void)
{
    pSourceSink object = (pSourceSink)malloc(sizeof(struct _SourceSink));
    if (object != NULL)
    {
        object->operation.action = ENCODE;
        object->operation.frameSize = 0;
        object->operation.blockSize = 0;
        object->operation.modeAndChannel = 0;
        object->read = &echoRead;
        object->write = &echoWrite;
    }
    return object;
}

/* ----------------------------------------------------------------------------
 * Function     : next
 * ----------------------------------------------------------------------------
 * Description  : the echo test data provider simply uses a predictable
 *                sequence of data to allow us to verify correct operation.
 *                This function is used to determine the next value in the
 *                sequence given a known value.
 * Inputs       : value : the current value
 * Outputs      : return the next value in sequence
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static uint32_t next(uint32_t value)
{
    return (value + 13) & 0xFF;
}

/* ----------------------------------------------------------------------------
 * Function     : echoRead
 * ----------------------------------------------------------------------------
 * Description  : Implementation of the read method to provide data for test
 * Inputs       : source : pointer to a source/sink control block
 *                buffer : the buffer which will be used to return the test
 *                data
 *                maxsize : The maximum size of the buffer which must not be
 *                exceeded
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void echoRead(pSourceSink source, unsigned char *buffer,
                     uint32_t maxsize)
{
    for (int i = 0; i < maxsize - 4; i++)
    {
        buffer[i] = (unsigned char)(readCounter & 0xFF);
        readCounter = next(readCounter);
    }
    source->srcUsed = maxsize - 4;
}

/* ----------------------------------------------------------------------------
 * Function     : echoWrite
 * ----------------------------------------------------------------------------
 * Description  : Implementation of the write method to verify test results
 * Inputs       : source : pointer to a source/sink control block
 *                buffer : the buffer which has been used to provide the
 *                test results
 *                size : The used size of the buffer
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void echoWrite(pSourceSink sink, unsigned char *buffer, uint32_t size)
{
    for (int i = 0; i < size; i++)
    {
        if (writeCounter != buffer[i])
        {
            logFatal(0xCCC);
        }
        writeCounter = next(writeCounter);
    }
    sink->snkUsed = size;
}
